@@ -1,13 +1,19 @@
 # dire.py
 
 """
-Provides the main class for dimensionality reduction.
+Provides the main JAX-based class for dimensionality reduction.
 
 The DiRe (Dimensionality Reduction) class implements a modern approach to
 dimensionality reduction, leveraging JAX for efficient computation. It uses 
 force-directed layout techniques combined with k-nearest neighbor graph 
 construction to generate meaningful low-dimensional embeddings of 
 high-dimensional data.
+
+This JAX implementation features:
+- Fully vectorized force computation (no chunking for optimal performance)
+- JIT compilation for mathematical operations
+- Optimized for small to medium datasets (<50K points)
+- Excellent CPU performance and research-friendly design
 """
 
 #
@@ -221,7 +227,7 @@ class DiRe(TransformerMixin):
         """ Column indices for nearest neighbors """
         self._adjacency = None
         """ kNN adjacency matrix """
-        self.random_state = None
+        self.random_state = random_state
         self.batch_size = batch_size
         #
         if my_logger is None:
@@ -545,14 +551,20 @@ class DiRe(TransformerMixin):
             key = random.PRNGKey(randint(0, 1000))
         else:
             key = random.PRNGKey(self.random_state)
-        rand_basis = random.normal(key, (self.n_components, self._data_dim))
+        
+        # Use appropriate dtype for MPA
+        compute_dtype = jnp.float32 if self.mpa else jnp.float64
+        rand_basis = random.normal(key, (self.n_components, self._data_dim), dtype=compute_dtype)
 
-        # Move data and projection matrix to device memory
-        data_matrix = device_put(self._data)
+        # Move data and projection matrix to device memory with consistent dtype
+        data_matrix = device_put(self._data.astype(compute_dtype))
         rand_basis = device_put(rand_basis)
 
-        # Project data onto random basis
-        self._init_embedding = data_matrix @ rand_basis.T
+        # Project data onto random basis with explicit precision control
+        self._init_embedding = jnp.dot(
+            data_matrix, rand_basis.T, 
+            precision=jax.lax.Precision.DEFAULT if self.mpa else jax.lax.Precision.HIGHEST
+        )
 
         self.logger.info("do_random_embedding done ...")
 
@@ -562,78 +574,20 @@ class DiRe(TransformerMixin):
 
     def do_rand_sampling(self, key, arr, n_samples, n_dirs, neg_ratio):
         """
-        Sample points for force calculation using random projections.
-
-        This method implements an efficient sampling strategy to identify points
-        for applying attractive and repulsive forces during layout optimization.
-        It uses random projections to quickly identify nearby points in different
-        directions, and also adds random negative samples for repulsion.
-
-        Parameters
-        ----------
-        key : jax.random.PRNGKey
-            Random number generator key
-        arr : jax.numpy.ndarray
-            Array of current point positions
-        n_samples : int
-            Number of samples to take in each direction
-        n_dirs : int
-            Number of random directions to sample
-        neg_ratio : int
-            Ratio of negative samples to positive samples
-
-        Returns
-        -------
-        jax.numpy.ndarray
-            Array of sampled indices for force calculations
-
-        Notes
-        -----
-        The sampling strategy works as follows:
-        1. Generate n_dirs random unit vectors
-        2. Project the points onto each vector
-        3. For each point, take the n_samples closest points in each direction
-        4. Add random negative samples for repulsion
-        5. Combine all sampled indices
-
-        This approach is more efficient than a full nearest neighbor search
-        while still capturing the important local relationships.
+        Sample points for force calculation using cached sampling kernel.
+        Uses a cached kernel to avoid recompilation issues with dynamic vmap.
         """
         self.logger.info("do_rand_sampling ...")
 
-        sampled_indices_list = []
         arr_len = len(arr)
+        compute_dtype = jnp.float32 if hasattr(self, 'mpa') and self.mpa else jnp.float64
 
-        # Get random unit vectors for projections
-        key, subkey = random.split(key)
-        direction_vectors = rand_directions(subkey, self.n_components, n_dirs)
+        # Use cached sampling kernel
+        sampling_kernel = _get_sampling_kernel(
+            self.n_components, n_samples, n_dirs, neg_ratio, compute_dtype
+        )
 
-        # For each direction, sample points based on projections
-        for vec in direction_vectors:
-            # Project points onto the direction vector
-            arr_proj = vec @ arr.T
-
-            # Sort indices by projection values
-            indices_sort = jnp.argsort(arr_proj)
-
-            # For each point, take n_samples points around it in sorted order
-            vmap_get_slice = vmap(get_slice, in_axes=(None, None, 0))
-            indices = vmap_get_slice(indices_sort, n_samples, jnp.arange(arr_len))
-
-            # Reorder indices back to original ordering
-            indices = indices[indices_sort]
-
-            # Add to list of sampled indices
-            sampled_indices_list.append(indices)
-
-        # Generate random negative samples for repulsion
-        n_neg_samples = int(neg_ratio * n_samples)
-        key, subkey = random.split(key)
-        neg_indices = random.randint(subkey, (arr_len, n_neg_samples), 0, arr_len)
-        sampled_indices_list.append(neg_indices)
-
-        # Combine all sampled indices
-        sampled_indices = jnp.concatenate(sampled_indices_list, axis=-1)
+        sampled_indices = sampling_kernel(key, arr)
 
         self.logger.info("do_rand_sampling done ...")
 
@@ -668,12 +622,13 @@ class DiRe(TransformerMixin):
         """
         self.logger.info("do_layout ...")
 
-        # Setup parameters
-        cutoff = jnp.array([self.cutoff])
-        num_iterations = self.max_iter_layout
+        # Setup parameters with appropriate dtype
+        compute_dtype = jnp.float32 if self.mpa else jnp.float64
 
         # Handle automatic batch size calculation if needed
         sample_size = self.sample_size
+        num_iterations = self.max_iter_layout
+
         if sample_size == "auto":
             # Scale batch size based on dataset size and neighborhood size
             sample_size = int(self.n_neighbors * np.log(self._n_samples))
@@ -695,21 +650,23 @@ class DiRe(TransformerMixin):
             f"[LAYOUT] Initial embedding precision: {self._init_embedding.dtype}"
         )
 
-        # we shall use force_cpu only as a flag passed to the routine
-        # force_cpu = force_cpu or large_dataset_mode and (jax.devices()[0].platform == 'tpu')
-
-        # Initialize and normalize positions
+        # Initialize and normalize positions with appropriate dtype
         if force_cpu:
             self.logger.info("Forcing computations on CPU")
             cpu_device = jax.devices("cpu")[0]
-            init_pos_jax = device_put(self._init_embedding, device=cpu_device)
+            init_pos_jax = device_put(self._init_embedding.astype(compute_dtype), device=cpu_device)
             neighbor_indices_jax = device_put(self._indices_np, device=cpu_device)
         else:
-            init_pos_jax = device_put(self._init_embedding)
+            init_pos_jax = device_put(self._init_embedding.astype(compute_dtype))
             neighbor_indices_jax = device_put(self._indices_jax)
 
-        init_pos_jax -= init_pos_jax.mean(axis=0)  # Center positions
-        init_pos_jax /= init_pos_jax.std(axis=0)  # Normalize variance
+        # Use dtype-consistent operations for normalization
+        mean_pos = init_pos_jax.mean(axis=0, keepdims=True)
+        init_pos_jax = init_pos_jax - mean_pos  # Center positions
+        std_pos = init_pos_jax.std(axis=0, keepdims=True)
+        # Avoid division by zero with appropriate epsilon for dtype
+        eps = jnp.array(1e-7 if self.mpa else 1e-15, dtype=compute_dtype)
+        init_pos_jax = init_pos_jax / jnp.maximum(std_pos, eps)  # Normalize variance
 
         # Set random seed for reproducibility
         if self.random_state is None:
@@ -717,85 +674,43 @@ class DiRe(TransformerMixin):
         else:
             key = random.PRNGKey(self.random_state)
 
-        # Optimization loop
-        for iter_id in tqdm(range(num_iterations)):
-            logger.debug(f"Iteration {iter_id + 1}")
+        # Use cached layout optimization kernel - this replaces the entire Python loop
+        # with a single JIT-compiled operation that avoids all recompilation issues
+        self.logger.info(f"Using cached layout optimization kernel for {num_iterations} iterations")
 
-            # Sample random points for repulsion
-            sample_indices_jax = self.do_rand_sampling(
-                key, init_pos_jax, sample_size, n_dirs, neg_ratio
-            )
+        # Use a reasonable max_iterations to allow caching across different iteration counts
+        # This enables a single kernel to handle 8, 16, 32, 64, 128 iterations efficiently
+        max_iterations = max(128, num_iterations)
 
-            if force_cpu:
-                cpu_device = jax.devices("cpu")[0]
-                sample_indices_jax = device_put(sample_indices_jax, device=cpu_device)
-            else:
-                sample_indices_jax = device_put(sample_indices_jax)
+        # Calculate chunk size for large dataset mode
+        chunk_size = None
+        if large_dataset_mode:
+            platform = jax.devices()[0].platform
+            base_chunk_size = self.memm.get(platform, self.memm.get("other", 8192))
+            # Reduce chunk size significantly for large dataset mode
+            chunk_size = max(base_chunk_size // 4, 1024)  # At least 1024, but 1/4 of normal batch size
+            self.logger.info(f"Using large dataset mode with chunk size: {chunk_size} (platform: {platform})")
+        else:
+            self.logger.info("Using standard dataset mode (no chunking)")
 
-            # Split computation for memory efficiency if needed
-            if large_dataset_mode:
-                # Process in chunks to reduce peak memory usage
-                if jax.devices()[0].platform == "tpu":
-                    chunk_size = min(self.memm["tpu"], self._n_samples)
-                elif jax.devices()[0].platform == "gpu":
-                    chunk_size = min(self.memm["gpu"], self._n_samples)
-                else:
-                    chunk_size = min(self.memm["other"], self._n_samples)
-                    # this is actually inefficient, but let's postpone
+        layout_kernel = _get_layout_optimization_kernel(
+            max_iterations=max_iterations,
+            sample_size=sample_size,
+            n_dirs=n_dirs,
+            neg_ratio=neg_ratio,
+            cutoff=self.cutoff,
+            a=self._a,
+            b=self._b,
+            use_float32=self.mpa,
+            large_dataset_mode=large_dataset_mode,
+            chunk_size=chunk_size
+        )
 
-                all_forces = []
-
-                self.logger.info(f"Using memory tiling with tile size: {chunk_size}")
-
-                for chunk_start in range(0, self._n_samples, chunk_size):
-                    chunk_end = min(chunk_start + chunk_size, self._n_samples)
-                    chunk_indices = jnp.arange(chunk_start, chunk_end)
-
-                    # Process this chunk using our kernelized function
-                    chunk_force = self._compute_forces(
-                        init_pos_jax,
-                        chunk_indices,
-                        neighbor_indices_jax[chunk_indices],
-                        sample_indices_jax[chunk_indices],
-                        alpha=1.0 - iter_id / num_iterations,
-                    )
-
-                    all_forces.append(chunk_force)
-
-                    # Explicitly clean up to reduce memory pressure
-                    gc.collect()
-
-                # Combine results from all chunks
-                net_force = jnp.concatenate(all_forces, axis=0)
-
-            else:
-                # Process all points at once for smaller datasets
-                net_force = self._compute_forces(
-                    init_pos_jax,
-                    jnp.arange(self._n_samples),
-                    neighbor_indices_jax,
-                    sample_indices_jax,
-                    alpha=1.0 - iter_id / num_iterations,
-                )
-
-            # Clip forces to prevent extreme movements
-            net_force = jnp.clip(net_force, -cutoff, cutoff)
-
-            # Update positions
-            init_pos_jax += net_force
-
-            # Ensure we're not accumulating unnecessary computation graphs in JAX
-            init_pos_jax.block_until_ready()
-
-            # Clean up resources
-            gc.collect()
-
-        # Normalize final layout
-        init_pos_jax -= init_pos_jax.mean(axis=0)
-        init_pos_jax /= init_pos_jax.std(axis=0)
+        # Run the complete optimization in a single JIT-compiled call with dynamic iterations
+        final_positions = layout_kernel(init_pos_jax, neighbor_indices_jax, key, num_iterations)
 
         # Store final layout
-        self._layout = np.asarray(init_pos_jax)
+        self._layout = np.asarray(final_positions)
 
         # Clear any cached values to free memory
         gc.collect()
@@ -831,15 +746,10 @@ class DiRe(TransformerMixin):
             Net force vectors for each point
         """
 
-        if self.mpa:
-            positions = positions.astype(jnp.float32)
-        else:
-            positions = positions.astype(jnp.float64)
-
         self.logger.debug(f"[FORCE] Computing forces on device: {positions.device}")
-        self.logger.debug(f"[FORCE] Using precision: {positions.dtype}")
+        self.logger.debug(f"[FORCE] Using MPA: {self.mpa}")
 
-        # Call the JAX-optimized kernel
+        # Call the JAX-optimized kernel with MPA flag
         return compute_forces_kernel(
             positions,
             chunk_indices,
@@ -848,6 +758,7 @@ class DiRe(TransformerMixin):
             alpha,
             self._a,
             self._b,
+            use_float32=self.mpa,
         )
 
     #
@@ -984,84 +895,88 @@ class DiRe(TransformerMixin):
 #
 
 
-@functools.partial(jit, static_argnums=(5, 6))
+# Cache force computation kernels to avoid recompilation
+@functools.lru_cache(maxsize=8)
+def _get_force_computation_kernel(a, b, use_float32):
+    """Get or create a cached force computation kernel for the given parameters."""
+
+    @jax.jit
+    def force_kernel(positions, chunk_indices, neighbor_indices, sample_indices, alpha):
+        return _compute_forces_impl(positions, chunk_indices, neighbor_indices, sample_indices, alpha, a, b, use_float32)
+
+    return force_kernel
+
 def compute_forces_kernel(
-    positions, chunk_indices, neighbor_indices, sample_indices, alpha, a, b
+    positions, chunk_indices, neighbor_indices, sample_indices, alpha, a, b, use_float32=True
 ):
     """
-    JAX-optimized kernel for computing attractive and repulsive forces.
-
-    Parameters
-    ----------
-    positions : jax.numpy.ndarray
-        Current point positions
-    chunk_indices : jax.numpy.ndarray
-        Current batch indices
-    neighbor_indices : jax.numpy.ndarray
-        Indices of neighbors for attractive forces
-    sample_indices : jax.numpy.ndarray
-        Indices of points for repulsive forces
-    alpha : float
-        Cooling factor that scales force magnitude
-    a : float
-        Attraction parameter
-    b : float
-        Repulsion parameter
-
-    Returns
-    -------
-    jax.numpy.ndarray
-        Net force vectors for each point
+    Cached wrapper for force computation to avoid recompilation.
+    Gets or creates a cached kernel for the given a, b, use_float32 configuration.
     """
+    kernel = _get_force_computation_kernel(a, b, use_float32)
+    return kernel(positions, chunk_indices, neighbor_indices, sample_indices, alpha)
 
-    # ===== Attraction Forces =====
-    def compute_attraction(chunk_idx, neighbors_idx):
-        # Get positions of current point and its neighbors
-        point_pos = positions[chunk_idx]
-        neighbor_pos = positions[neighbors_idx]
+def _compute_forces_impl(
+    positions, chunk_indices, neighbor_indices, sample_indices, alpha, a, b, use_float32
+):
+    """
+    Fully vectorized JAX implementation of force computation.
+    This is the core computation that gets JIT-compiled and cached.
+    """
+    
+    # Convert to appropriate dtype at the beginning
+    compute_dtype = jnp.float32 if use_float32 else jnp.float64
+    positions = positions.astype(compute_dtype)
+    alpha = jnp.array(alpha, dtype=compute_dtype)
+    a = jnp.array(a, dtype=compute_dtype)
+    b = jnp.array(b, dtype=compute_dtype)
+    
+    # Get current positions for processing
+    current_positions = positions[chunk_indices]  # (N, D)
+    forces = jnp.zeros_like(current_positions, dtype=compute_dtype)
 
-        # Compute position differences and distances
-        diff = neighbor_pos - point_pos
-        dist = jnp.linalg.norm(diff, axis=1, keepdims=True)
+    # ===== ATTRACTION FORCES (k-NN only) =====
+    # Get neighbor positions efficiently using advanced indexing
+    neighbor_positions = positions[neighbor_indices]  # (N, k, D)
+    
+    # Broadcast current positions for vectorized computation
+    current_pos_expanded = current_positions[:, None, :]  # (N, 1, D)
+    
+    # Compute differences and distances
+    att_diff = neighbor_positions - current_pos_expanded  # (N, k, D)
+    # Use more stable distance computation for MPA
+    att_dist_sq = jnp.sum(att_diff * att_diff, axis=2, keepdims=True)  # (N, k, 1)
+    att_dist = jnp.sqrt(att_dist_sq + jnp.array(1e-10, dtype=compute_dtype))  # (N, k, 1)
+    
+    # Attraction coefficient: 1 / (1 + a * (1/d)^(2b))
+    # Use more numerically stable computation
+    inv_dist = jnp.array(1.0, dtype=compute_dtype) / att_dist
+    att_coeff = jnp.array(1.0, dtype=compute_dtype) / (jnp.array(1.0, dtype=compute_dtype) + a * jnp.power(inv_dist, 2.0 * b))  # (N, k, 1)
+    
+    # Compute attraction forces and sum over neighbors
+    att_forces = jnp.sum(att_coeff * att_diff / att_dist, axis=1)  # (N, D)
+    forces += att_forces
 
-        # Avoid division by zero
-        mask = dist > 0
-        direction = jnp.where(mask, diff / dist, 0.0)
+    # ===== REPULSION FORCES (Random Sampling) =====
+    if sample_indices.size > 0:
+        # Get negative sample positions
+        sample_positions = positions[sample_indices]  # (N, n_neg, D)
+        
+        # Compute differences and distances
+        rep_diff = sample_positions - current_pos_expanded  # (N, n_neg, D)
+        # Use more stable distance computation for MPA
+        rep_dist_sq = jnp.sum(rep_diff * rep_diff, axis=2, keepdims=True)  # (N, n_neg, 1)
+        rep_dist = jnp.sqrt(rep_dist_sq + jnp.array(1e-10, dtype=compute_dtype))  # (N, n_neg, 1)
+        
+        # Repulsion coefficient: -1 / (1 + a * d^(2b))
+        rep_coeff = jnp.array(-1.0, dtype=compute_dtype) / (jnp.array(1.0, dtype=compute_dtype) + a * jnp.power(rep_dist, 2.0 * b))  # (N, n_neg, 1)
+        
+        # Compute repulsion forces and sum over negative samples
+        rep_forces = jnp.sum(rep_coeff * rep_diff / rep_dist, axis=1)  # (N, D)
+        forces += rep_forces
 
-        # Compute attraction-repulsion coefficients
-        grad_coeff = jnp.where(
-            mask, 1.0 * jax_coeff_att(dist, a, b) + 1.0 * jax_coeff_rep(dist, a, b), 0.0
-        )
-
-        # Sum forces from all neighbors
-        return jnp.sum(grad_coeff * direction, axis=0)
-
-    # ===== Repulsion Forces =====
-    def compute_repulsion(chunk_idx, sample_idx):
-        # Get positions of current point and sampled points
-        point_pos = positions[chunk_idx]
-        sample_pos = positions[sample_idx]
-
-        # Compute position differences and distances
-        diff = sample_pos - point_pos
-        dist = jnp.linalg.norm(diff, axis=1, keepdims=True)
-
-        # Avoid division by zero
-        mask = dist > 0
-        direction = jnp.where(mask, diff / dist, 0.0)
-
-        # Compute repulsion coefficients
-        grad_coeff = jnp.where(mask, jax_coeff_rep(dist, a, b), 0.0)
-
-        # Sum forces from all sampled points
-        return jnp.sum(grad_coeff * direction, axis=0)
-
-    # Vectorize force computation across all points
-    attraction_forces = vmap(compute_attraction)(chunk_indices, neighbor_indices)
-    repulsion_forces = vmap(compute_repulsion)(chunk_indices, sample_indices)
-
-    # Combine forces with cooling factor
-    return alpha * (attraction_forces + repulsion_forces)
+    # Apply cooling factor
+    return alpha * forces
 
 
 #
@@ -1069,10 +984,20 @@ def compute_forces_kernel(
 #
 
 
-@jax.jit
+# Cache coefficient kernels to avoid recompilation
+@functools.lru_cache(maxsize=8)
+def _get_distribution_kernel(a, b):
+    """Get or create a cached distribution kernel for the given parameters."""
+
+    @jax.jit
+    def dist_kernel(dist):
+        return 1.0 / (1.0 + a * dist ** (2 * b))
+
+    return dist_kernel
+
 def distribution_kernel(dist, a, b):
     """
-    Probability kernel that maps distances to similarity scores.
+    Probability kernel that maps distances to similarity scores using cached kernels.
 
     This is a rational function approximation of a t-distribution.
 
@@ -1090,26 +1015,59 @@ def distribution_kernel(dist, a, b):
     float or jax.numpy.ndarray
         Similarity score(s) between 0 and 1
     """
-    return 1.0 / (1.0 + a * dist ** (2 * b))
+    kernel = _get_distribution_kernel(a, b)
+    return kernel(dist)
 
+# Cache coefficient kernels
+@functools.lru_cache(maxsize=8)
+def _get_coeff_att_kernel(a, b):
+    """Get or create a cached attraction coefficient kernel."""
+
+    @jax.jit
+    def att_kernel(dist):
+        return 1.0 / (1.0 + a * (1.0 / dist) ** (2 * b))
+
+    return att_kernel
+
+@functools.lru_cache(maxsize=8)
+def _get_coeff_rep_kernel(a, b):
+    """Get or create a cached repulsion coefficient kernel."""
+
+    @jax.jit
+    def rep_kernel(dist):
+        return -1.0 / (1.0 + a * dist ** (2 * b))
+
+    return rep_kernel
 
-# Helper functions for force calculations
-@jax.jit
 def jax_coeff_att(dist, a, b):
-    """JAX-optimized attraction coefficient function."""
-    return 1.0 * distribution_kernel(1 / dist, a, b)
+    """JAX-optimized attraction coefficient function with caching."""
+    kernel = _get_coeff_att_kernel(a, b)
+    return kernel(dist)
 
-
-@jax.jit
 def jax_coeff_rep(dist, a, b):
-    """JAX-optimized repulsion coefficient function."""
-    return -1.0 * distribution_kernel(dist, a, b)
+    """JAX-optimized repulsion coefficient function with caching."""
+    kernel = _get_coeff_rep_kernel(a, b)
+    return kernel(dist)
 
 
-@functools.partial(jit, static_argnums=(1, 2))
-def rand_directions(key, dim=2, num=100):
+# Cache random direction kernels to avoid recompilation
+@functools.lru_cache(maxsize=8)
+def _get_rand_directions_kernel(dim, num, dtype):
+    """Get or create a cached random directions kernel for the given configuration."""
+
+    @jax.jit
+    def rand_dir_kernel(key):
+        points = random.normal(key, (num, dim), dtype=dtype)
+        norms = jnp.sqrt(jnp.sum(points * points, axis=-1, keepdims=True))
+        # Add small epsilon to avoid division by zero
+        eps = jnp.array(1e-7 if dtype == jnp.float32 else 1e-15, dtype=dtype)
+        return points / jnp.maximum(norms, eps)
+
+    return rand_dir_kernel
+
+def rand_directions(key, dim=2, num=100, dtype=jnp.float64):
     """
-    Sample unit vectors in random directions.
+    Sample unit vectors in random directions using cached kernels.
 
     Parameters
     ----------
@@ -1119,21 +1077,32 @@ def rand_directions(key, dim=2, num=100):
         Dimensionality of the vectors
     num : int
         Number of random directions to sample
+    dtype : jnp.dtype
+        Data type for computation
 
     Returns
     -------
     jax.numpy.ndarray
         Array of shape (num, dim) containing unit vectors
     """
-    points = random.normal(key, (num, dim))
-    norms = jnp.sqrt(jnp.sum(points * points, axis=-1))
-    return points / norms[:, None]
+    kernel = _get_rand_directions_kernel(dim, num, dtype)
+    return kernel(key)
 
 
-@functools.partial(jit, static_argnums=(1,))
+# Cache slice kernels to avoid recompilation
+@functools.lru_cache(maxsize=16)
+def _get_slice_kernel(k):
+    """Get or create a cached slice kernel for the given slice size."""
+
+    @jax.jit
+    def slice_kernel(arr, i):
+        return lax.dynamic_slice(arr, (i - k // 2,), (k,))
+
+    return slice_kernel
+
 def get_slice(arr, k, i):
     """
-    Extract a slice of size k centered around index i.
+    Extract a slice of size k centered around index i using cached kernels.
 
     Parameters
     ----------
@@ -1149,4 +1118,202 @@ def get_slice(arr, k, i):
     jax.numpy.ndarray
         Slice of the input array
     """
-    return lax.dynamic_slice(arr, (i - k // 2,), (k,))
+    kernel = _get_slice_kernel(k)
+    return kernel(arr, i)
+
+# Cache sampling kernels to avoid recompilation in do_rand_sampling
+@functools.lru_cache(maxsize=16)
+def _get_sampling_kernel(n_components, n_samples, n_dirs, neg_ratio, dtype):
+    """Get or create a cached sampling kernel for the given configuration."""
+
+    @jax.jit
+    def sampling_kernel(key, arr):
+        arr_len = arr.shape[0]
+        sampled_indices_list = []
+
+        # Get random unit vectors for projections
+        key, subkey = random.split(key)
+        direction_vectors = _get_rand_directions_kernel(n_components, n_dirs, dtype)(subkey)
+
+        # Get slice kernel for this n_samples
+        slice_kernel = _get_slice_kernel(n_samples)
+
+        # For each direction, sample points based on projections
+        for i in range(n_dirs):
+            vec = direction_vectors[i]
+
+            # Project points onto the direction vector
+            arr_proj = jnp.dot(vec, arr.T)
+
+            # Sort indices by projection values
+            indices_sort = jnp.argsort(arr_proj)
+
+            # For each point, take n_samples points around it in sorted order
+            # Use vmap on the cached slice kernel
+            vmap_slice = vmap(slice_kernel, in_axes=(None, 0))
+            indices = vmap_slice(indices_sort, jnp.arange(arr_len))
+
+            # Reorder indices back to original ordering
+            indices = indices[indices_sort]
+
+            # Add to list of sampled indices
+            sampled_indices_list.append(indices)
+
+        # Generate random negative samples for repulsion
+        n_neg_samples = int(neg_ratio * n_samples)
+        key, subkey = random.split(key)
+        neg_indices = random.randint(subkey, (arr_len, n_neg_samples), 0, arr_len)
+        sampled_indices_list.append(neg_indices)
+
+        # Combine all sampled indices
+        sampled_indices = jnp.concatenate(sampled_indices_list, axis=-1)
+
+        return sampled_indices
+
+    return sampling_kernel
+
+
+def _get_chunked_sampling_kernel(n_components, n_samples, n_dirs, neg_ratio, dtype, chunk_size):
+    """Get or create a memory-efficient chunked sampling kernel for large datasets."""
+
+    @jax.jit
+    def chunked_sampling_kernel(key, arr):
+        arr_len = arr.shape[0]
+
+        # For large datasets, use a simpler approximation that reduces memory usage
+        # Instead of processing in dynamic chunks, we use subsampling to reduce the problem size
+
+        # Get random unit vectors for projections
+        key, subkey = random.split(key)
+        direction_vectors = _get_rand_directions_kernel(n_components, n_dirs, dtype)(subkey)
+
+        # Subsample the array to reduce memory footprint
+        subsample_size = min(chunk_size, arr_len)
+        key, subkey = random.split(key)
+        subsample_indices = random.choice(subkey, arr_len, (subsample_size,), replace=False)
+        subsample_arr = arr[subsample_indices]
+
+        sampled_indices_list = []
+
+        # For each direction, sample points based on projections from subsampled array
+        for i in range(n_dirs):
+            vec = direction_vectors[i]
+
+            # Project subsampled points onto the direction vector
+            subsample_proj = jnp.dot(vec, subsample_arr.T)
+
+            # Sort indices by projection values
+            indices_sort = jnp.argsort(subsample_proj)
+
+            # For each original point, find nearest points in the subsampled space
+            arr_proj_full = jnp.dot(vec, arr.T)
+
+            def find_neighbors_for_point(point_proj):
+                # Find position in sorted subsampled projections
+                pos = jnp.searchsorted(subsample_proj[indices_sort], point_proj)
+
+                # Sample around this position in the subsampled space
+                start = jnp.maximum(0, pos - n_samples // 2)
+                end = jnp.minimum(subsample_size, start + n_samples)
+                start = jnp.maximum(0, end - n_samples)
+
+                # Map back to original indices using dynamic_slice
+                subsampled_neighbors = jax.lax.dynamic_slice(indices_sort, (start,), (n_samples,))
+                return subsample_indices[subsampled_neighbors]
+
+            vmap_find_neighbors = vmap(find_neighbors_for_point)
+            indices = vmap_find_neighbors(arr_proj_full)
+            sampled_indices_list.append(indices)
+
+        # Generate random negative samples
+        n_neg_samples = int(neg_ratio * n_samples)
+        key, subkey = random.split(key)
+        neg_indices = random.randint(subkey, (arr_len, n_neg_samples), 0, arr_len)
+        sampled_indices_list.append(neg_indices)
+
+        # Combine all sampled indices
+        sampled_indices = jnp.concatenate(sampled_indices_list, axis=-1)
+
+        return sampled_indices
+
+    return chunked_sampling_kernel
+
+# Cache complete layout optimization kernels with max iterations
+@functools.lru_cache(maxsize=8)
+def _get_layout_optimization_kernel(max_iterations, sample_size, n_dirs, neg_ratio, cutoff, a, b, use_float32, large_dataset_mode=False, chunk_size=None):
+    """Get or create a cached layout optimization kernel for the given configuration."""
+
+    @jax.jit
+    def layout_kernel(init_positions, neighbor_indices, initial_key, actual_iterations):
+        """
+        Complete JIT-compiled layout optimization loop with dynamic iteration count.
+        This replaces the entire Python for-loop with a JAX-optimized implementation.
+        """
+        n_components = init_positions.shape[1]
+        compute_dtype = jnp.float32 if use_float32 else jnp.float64
+
+        # Ensure consistent dtypes
+        positions = init_positions.astype(compute_dtype)
+        cutoff_typed = jnp.array(cutoff, dtype=compute_dtype)
+        a_typed = jnp.array(a, dtype=compute_dtype)
+        b_typed = jnp.array(b, dtype=compute_dtype)
+
+        # Get cached kernels - use chunked sampling for large datasets to reduce memory usage
+        if large_dataset_mode and chunk_size is not None:
+            sampling_kernel = _get_chunked_sampling_kernel(
+                n_components, sample_size, n_dirs, neg_ratio, compute_dtype, chunk_size
+            )
+        else:
+            sampling_kernel = _get_sampling_kernel(n_components, sample_size, n_dirs, neg_ratio, compute_dtype)
+
+        force_kernel = _get_force_computation_kernel(a, b, use_float32)
+
+        def optimization_step(carry, iter_id):
+            current_positions, key = carry
+
+            # Split key for this iteration
+            key, subkey = random.split(key)
+
+            # Skip iterations beyond actual_iterations
+            skip_iteration = iter_id >= actual_iterations
+
+            # Calculate cooling factor (use actual_iterations for proper scaling)
+            alpha = jnp.where(skip_iteration, 0.0, 1.0 - iter_id / actual_iterations)
+
+            # Sample points for force calculation
+            sample_indices = sampling_kernel(subkey, current_positions)
+
+            # Conditionally compute forces and update positions using lax.cond
+            def do_update(_):
+                n_points = current_positions.shape[0]
+                chunk_indices = jnp.arange(n_points)
+                net_force = force_kernel(
+                    current_positions, chunk_indices, neighbor_indices, sample_indices, alpha
+                )
+                net_force = jnp.clip(net_force, -cutoff_typed, cutoff_typed)
+                return current_positions + net_force
+
+            def skip_update(_):
+                return current_positions
+
+            new_positions = lax.cond(skip_iteration, skip_update, do_update, None)
+
+            return (new_positions, key), None
+
+        # Run the optimization loop using scan for efficiency with max iterations
+        (final_positions, _), _ = lax.scan(
+            optimization_step,
+            (positions, initial_key),
+            jnp.arange(max_iterations)
+        )
+
+        # Final normalization
+        mean_final = final_positions.mean(axis=0, keepdims=True)
+        final_positions = final_positions - mean_final
+        std_final = final_positions.std(axis=0, keepdims=True)
+        eps = jnp.array(1e-7 if use_float32 else 1e-15, dtype=compute_dtype)
+        final_positions = final_positions / jnp.maximum(std_final, eps)
+
+        return final_positions
+
+    return layout_kernel
