@@ -15,7 +15,7 @@ sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')
 def run_all_tests():
     """Run all unit tests for the DiRe package."""
     test_loader = unittest.TestLoader()
-    test_suite = test_loader.discover('tests/unit', pattern='test_*.py')
+    test_suite = test_loader.discover('unit', pattern='test_*.py')
     test_runner = unittest.TextTestRunner(verbosity=2)
     result = test_runner.run(test_suite)
     return result.wasSuccessful()
