@@ -3,6 +3,10 @@
 Simple test of DiRe with timing.
 """
 
+import sys
+import os
+sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
+
 import time
 
 from sklearn.datasets import make_blobs
